-- --------------------------------------------------------------------------
-- Unit tester for Lpm.bs
-- --------------------------------------------------------------------------

package TestLpm (mkTestLpm)
where

import FileIO
import Monad
import Printf
import FIFO
import ArrayFile

import GetPut
import ClientServer
import RAM
import SyncSRAM
import SRAM
import SRAMFile
import SPSRAM

import ROM

import LPMifc
import LPM
import LPMK

type LuTag = Bit 8

-- ----------------------------------------------------------------
-- Top-level test-jig module for testing using the C backend
-- - instantiates a fake SRAM (sriLpm) which loads its contents from
--       the file "sram_contents.txt"
-- - instantiates the longest-prefix-match module (lpm)
-- - repeatedly (until end of file "test.data")
--   - reads an IP address from file "test.data"
--   - sends the IP address and a tag into the lpm, and also enqueues
--         the current clock
--   - accepts a result from the lpm and prints it out with the clock tick

data State = Initializing | Running | Stopped
           deriving (Eq, Bits)

-- C tester
mkTestLpm :: Module Empty
mkTestLpm =
    module

        -- Two possible implementations of a basic RAM;
        -- comment out the version you don't want
	ram :: RAM LuAddr LuData
--	ram <- mkWrapSRAM (mkSRAMFile "sram_contents.txt" (exp 2 (valueOf (SizeOf LuAddr))))
	ram <- mkRAMFile "sram_contents.txt" (exp 2 (valueOf (SizeOf LuAddr)))
	let rom :: AsyncROM 1 LuAddr LuData
	    rom = convRAMToROM ram

        -- Two possible implementations of the longest prefix match module;
        -- comment out the version you don't want
        lpm :: LPM
        lpm <- mkLPM rom           -- The straight pipeline version
--        lpm <- mkLPMK rom          -- The circular pipeline version

        file :: FileGet (Bit 32)
        file <- mkFileHexGet  "test.data"

        state :: Reg State
        state <- mkReg Initializing

	ticks :: Reg (Bit 32) <- mkReg 0

        rules
          "TestLpm init":
            when (state == Initializing)
             ==> action
                  state := Running

          "TestLpm read":
            when (state == Running)
             ==> action
                  m_ipa :: Maybe (Bit 32) <- file.get

                  case m_ipa of
                    Nothing  -> state := Stopped
                    Just ipa -> action
                                  lpm.luReq ipa
                                  printf "%4d: File data ipa 0x%08x\n"
					 (arg ticks)
                                         (arg ipa)

          "TestLpm result":
            when res <- lpm.luResp
             ==> action
		  lpm.luRespAck
	          printf "%4d: Result 0x%08x\n"
			 (arg ticks)
                         (arg res)

          "TestLpm tick":
            when True
             ==> ticks := ticks + 1

------------
-- Used by testjig above for C backend

-- Fake RAM for testing
mkRAMFile :: (Bits a sa, Bits d sd) => String -> Integer -> Module (RAM a d)
mkRAMFile file size =
    module
	buf :: FIFO d <- mkSizedFIFO 1
	arr :: Array a d <- mkArrayFile file (unpack 0) (unpack (fromInteger (size - 1)))
	interface Server
	    request =
             interface Put
              put (Read a) = buf.enq (arr.sub a)
	      put (Write d a) = arr.upd a d
	    response =
             interface Get
              get = do
		buf.deq
		return buf.first

-- --------------------------------------------------------------
-- Alternative top-level module for Verilog back end.
-- Contains an "embedded" SRAM

-- internal SRAM
mkLpmI :: Module LPM
mkLpmI =
    module
	ram :: RAM LuAddr LuData
	ram <- mkWrapSRAM (mkSPSRAM (exp 2 (valueOf (SizeOf LuAddr))))
	let rom :: AsyncROM 5 LuAddr LuData
	    rom = convRAMToROM ram
	mkLPM rom

-- --------------------------------------------------------------
-- Alternative top-level module for Verilog back end.
-- Presents an interface to an external SRAM

-- external SRAM
-- XXX This doesn't work yet. {-# verilog mkLpmE{noReady,alwaysEnabled} #-}
mkLpmE :: Module (SyncSRAMC 1 (SizeOf LuAddr) (SizeOf LuData), LPM)
mkLpmE =
    module
	(extram, ram) <- wrapSRAM
	let rom :: AsyncROM 5 LuAddr LuData
	    rom = convRAMToROM ram
	lpm <- mkLPM rom
	interface (extram, lpm)
