/* -*-	Mode:C; c-basic-offset:4 -*- */

/* --------------------------------------------------------------------------
 * This program displays the SRAM produced by the table generator for
 * IPv4 Longest Prefix Match.
 *
 * See "print_usage()" for input file formats.
 * 
 * The table has the following simple structure:
 *   Top-level is a 2^DEPTH1 array, indexed by the first DEPTH1 bits
 *     of the IP address being looked up.
 *     Each entry is either a LEAF (MSB=1) or NONLEAF (MSB=0)
 *     If LEAF, remaining 31b contains Lookup Result
 *     If NONLEAF, points to a CHUNK
 *
 *   Subsequent levels are CHUNKs.  Each CHUNK has 2^DEPTH2 entries,
 *   indexed by the next DEPTH2 bits of the IP lookup address.
 *   Each entry is either a LEAF, or NONLEAF (in which case
 *   it points to the next-level CHUNK.
 *
 * Nikhil, August 25, 2002
 * --------------------------------------------------------------------------
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* ================================================================
 * Some general defs (not specific to this application)
 */

#define TRUE    1
#define FALSE   0

#define eqString(s1,s2)  (strcmp((s1),(s2)) == 0)

typedef  unsigned int  uint32_t;

/* ================================================================
 * Tree defs
 */

#define DEPTH1  16
/* #define DEPTH2   4 */
#define DEPTH2   8

#define LEAFTAG           0x80000000
#define isLeaf(x)         (((x) & LEAFTAG) != 0)
#define mkLeaf(x)         ((x) | LEAFTAG)
#define leafContents(x)   ((x) & (~ LEAFTAG))
#define defaultLeaf       mkLeaf(0)

/* ================================================================
 * The SRAM
 */

#define  SRAM_NUM_ADDR_BITS   21
#define  SRAM_SIZE            (1 << SRAM_NUM_ADDR_BITS)
#define  SRAM_ADDR_MASK       (SRAM_SIZE - 1)

static
uint32_t  sram [SRAM_SIZE];

/* ================================================================
 * lookupChunk()
 */

static
uint32_t
lookupChunk (uint32_t  ipa,
	     uint32_t  baseAddr,
	     uint32_t  depth)
{
    uint32_t  addr, val;

    addr = baseAddr + (ipa >> (32 - depth));
    val = sram [addr];

    printf ("sram [0x%05x] ==> 0x%08x\n", addr, val);

    if (isLeaf (val))
	return val;
    else
	return lookupChunk (ipa << depth, val, DEPTH2);
} /* lookupChunk() */

/* ****************************************************************
 * main() and friends
 */

/* ----------------------------------------------------------------
 * print_usage()
 */

static
void
print_usage (FILE *fp, int argc, char *argv[])
{
    fprintf (fp, "Usage:    %s  <sram input filename>  <ipas input filename\n", argv[0]);
    fprintf (fp, "\n");
    fprintf (fp, "<input filename>        Contains lookup table memory map. Each line has:\n");
    fprintf( fp, "                           @ <addr> <32b value>\n");
    fprintf (fp, "\n");
    fprintf (fp, "<ipas input filename>   Contains ip addresses. Each line has:\n");
    fprintf( fp, "                            <IP addr>\n");
} /* print_usage() */

/* ----------------------------------------------------------------
 * main()
 */

#define LINESIZE  1024

int
main (int argc, char *argv[])
{
    FILE       *fpSram, *fpIpas;
    uint32_t   addr, val, ipa, result;
    int        lineNum, n;
    char       inputLine [LINESIZE], *s;

    if ((argc > 1) && (eqString (argv[1], "?") ||
		       eqString (argv[1], "-h") ||
		       eqString (argv[1], "-help") ||
		       eqString (argv[1], "--help"))) {
	print_usage (stdout, argc, argv);
	return 0;
    }

    if (argc != 3) {
	print_usage (stderr, argc, argv);
	return 1;
    }

    fpSram = fopen (argv [1], "r");
    if (fpSram == NULL) {
	fprintf (stderr, "***** Error: couldn't open input sram file: %s\n", argv [1]);
	return 1;
    }

    lineNum       = 0;
    while (! (feof (fpSram))) {
	s = fgets (inputLine, LINESIZE, fpSram);
	if (s == NULL) break;
	lineNum++;

	n = sscanf (inputLine, "@%x %x", & addr, & val);

	/* ---- Check inputs */
	if (n != 2) {
	    inputLine [strlen (inputLine) - 1] = 0;
	    fprintf (stderr, "***** Error on line %d: \"%s\"\n", lineNum, inputLine);
	    fprintf (stderr, "***** Syntax error; couldn't parse two integers\n");
	    fprintf (stderr, "***** Aborting ...\n");
	    return 1;
	}

	if (addr >= SRAM_SIZE) {
	    inputLine [strlen (inputLine) - 1] = 0;
	    fprintf (stderr, "***** Error on line %d: \"%s\"\n", lineNum, inputLine);
	    fprintf (stderr, "***** Data error: addr (0x%08x) > SRAM_SIZE (0x%05x\n",
		     addr, SRAM_SIZE);
	    fprintf (stderr, "***** Aborting ...\n");
	    return 1;
	}

	sram [addr] = val;
    }

    fclose (fpSram);

    fpIpas = fopen (argv [2], "r");
    if (fpIpas == NULL) {
	fprintf (stderr, "***** Error: couldn't open input ip addrs file: %s\n", argv [2]);
	return 1;
    }

    lineNum       = 0;
    while (! (feof (fpIpas))) {
	s = fgets (inputLine, LINESIZE, fpIpas);
	if (s == NULL) break;
	lineNum++;

	n = sscanf (inputLine, "%x", & ipa);

	/* ---- Check input */
	if (n != 1) {
	    inputLine [strlen (inputLine) - 1] = 0;
	    fprintf (stderr, "***** Error on line %d: \"%s\"\n", lineNum, inputLine);
	    fprintf (stderr, "***** Syntax error; couldn't parse ip address\n");
	    fprintf (stderr, "***** Ignoring this line\n");
	    continue;
	}
	printf ("Lookin up 0x%08x\n", ipa);
	result = lookupChunk (ipa, 0, DEPTH1);
	printf ("Lookin up 0x%08x: result is 0x%08x\n", ipa, result);
    }

    return 0;
} /* main() */

/* **************************************************************** */
