/* -*-	Mode:C; c-basic-offset:4 -*- */

/* --------------------------------------------------------------------------
 * This program displays the SRAM produced by the table generator for
 * IPv4 Longest Prefix Match.
 *
 * See "print_usage()" for input file formats.
 * 
 * The table has the following simple structure:
 *   Top-level is a 2^DEPTH1 array, indexed by the first DEPTH1 bits
 *     of the IP address being looked up.
 *     Each entry is either a LEAF (MSB=1) or NONLEAF (MSB=0)
 *     If LEAF, remaining 31b contains Lookup Result
 *     If NONLEAF, points to a CHUNK
 *
 *   Subsequent levels are CHUNKs.  Each CHUNK has 2^DEPTH2 entries,
 *   indexed by the next DEPTH2 bits of the IP lookup address.
 *   Each entry is either a LEAF, or NONLEAF (in which case
 *   it points to the next-level CHUNK.
 *
 * Nikhil, August 25, 2002
 * --------------------------------------------------------------------------
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* ================================================================
 * Some general defs (not specific to this application)
 */

#define TRUE    1
#define FALSE   0

#define eqString(s1,s2)  (strcmp((s1),(s2)) == 0)

typedef  unsigned int  uint32_t;

/* ================================================================
 * Tree defs
 */

#define DEPTH1  16
/* #define DEPTH2   4 */
#define DEPTH2   8

#define LEAFTAG           0x80000000
#define isLeaf(x)         (((x) & LEAFTAG) != 0)
#define mkLeaf(x)         ((x) | LEAFTAG)
#define leafContents(x)   ((x) & (~ LEAFTAG))
#define defaultLeaf       mkLeaf(0)

/* ================================================================
 * The SRAM
 */

#define  SRAM_NUM_ADDR_BITS   21
#define  SRAM_SIZE            (1 << SRAM_NUM_ADDR_BITS)
#define  SRAM_ADDR_MASK       (SRAM_SIZE - 1)

static
uint32_t  sram [SRAM_SIZE];

/* ================================================================
 * Show the table
 */

/* ----------------------------------------------------------------
 * showPendingConsecutiveLeafs()
 */

static
void
showPendingConsecutiveLeafs (uint32_t  addr1,
			     uint32_t  addr2,
			     uint32_t  val)
{
    if (val != 0) {
	if (addr1 < addr2)
	    printf ("0x%05x ...\n", addr1);
	printf ("0x%05x: 0x%08x\n", addr2, val);
    }
} /* showPendingConsecutiveLeafs() */

/* ----------------------------------------------------------------
 * showChunk()
 */

static
void
showChunk (int       sliceNum,
	   uint32_t  parentAddr,
	   uint32_t  baseAddr,
	   uint32_t  depth)
{
    uint32_t  addr, val, startAddr, curVal;

    printf ("---------------- Slice %d chunk", sliceNum);
    if (parentAddr < SRAM_SIZE) printf (" (parent 0x%05x)", parentAddr);
    printf ("\n");

    curVal = 0;    /* ---- Bogus Non-leaf ptr to addr 0 */

    /* ---- Show chunk */
    for (addr = baseAddr; addr < (baseAddr + (1 << depth)); addr++) {
	val = sram [addr];
	if (! (isLeaf (val))) {
	    showPendingConsecutiveLeafs (startAddr, addr - 1, curVal);
	    printf ("0x%05x: --> Chunk at 0x%08x\n", addr, val);
	    curVal = 0;
	}
	else if (curVal == 0) {
	    curVal    = val;
	    startAddr = addr;
	}
	else if (val != curVal) {
	    showPendingConsecutiveLeafs (startAddr, addr - 1, curVal);
	    curVal    = val;
	    startAddr = addr;
	}
    }
    showPendingConsecutiveLeafs (startAddr, addr - 1, curVal);

    /* ---- Show chunks below */
    for (addr = baseAddr; addr < (baseAddr + (1 << depth)); addr++) {
	val = sram [addr];
	if (! (isLeaf (val))) {
	    showChunk (sliceNum + 1, addr, val, DEPTH2);
	}
    }
} /* showChunk() */

/* ****************************************************************
 * main() and friends
 */

/* ----------------------------------------------------------------
 * print_usage()
 */

static
void
print_usage (FILE *fp, int argc, char *argv[])
{
    fprintf (fp, "Usage:    %s  <input filename>\n", argv[0]);
    fprintf (fp, "\n");
    fprintf (fp, "<input filename>   Contains lookup table memory map. Each line has:\n");
    fprintf( fp, "                       <addr> <32b value>\n");
} /* print_usage() */

/* ----------------------------------------------------------------
 * main()
 */

#define LINESIZE  1024

int
main (int argc, char *argv[])
{
    FILE      *fpIn;
    uint32_t   addr, val;
    int        lineNum, n;
    char       inputLine [LINESIZE], *s;

    if ((argc > 1) && (eqString (argv[1], "?") ||
		       eqString (argv[1], "-h") ||
		       eqString (argv[1], "-help") ||
		       eqString (argv[1], "--help"))) {
	print_usage (stdout, argc, argv);
	return 0;
    }

    if (argc != 2) {
	print_usage (stderr, argc, argv);
	return 1;
    }

    fpIn = fopen (argv [1], "r");
    if (fpIn == NULL) {
	fprintf (stderr, "***** Error: couldn't open input file: %s\n", argv [1]);
	return 1;
    }

    lineNum       = 0;
    while (! (feof (fpIn))) {
	s = fgets (inputLine, LINESIZE, fpIn);
	if (s == NULL) break;
	lineNum++;

	n = sscanf (inputLine, "@%x %x", & addr, & val);

	/* ---- Check inputs */
	if (n != 2) {
	    inputLine [strlen (inputLine) - 1] = 0;
	    fprintf (stderr, "***** Error on line %d: \"%s\"\n", lineNum, inputLine);
	    fprintf (stderr, "***** Syntax error; couldn't parse two integers\n");
	    fprintf (stderr, "***** Aborting ...\n");
	    return 1;
	}

	if (addr >= SRAM_SIZE) {
	    inputLine [strlen (inputLine) - 1] = 0;
	    fprintf (stderr, "***** Error on line %d: \"%s\"\n", lineNum, inputLine);
	    fprintf (stderr, "***** Data error: addr (0x%08x) > SRAM_SIZE (0x%05x\n",
		     addr, SRAM_SIZE);
	    fprintf (stderr, "***** Aborting ...\n");
	    return 1;
	}

	sram [addr] = val;
    }

    showChunk (0, 0xFFFFFFFF, 0, DEPTH1);

    fclose (fpIn);

    return 0;
} /* main() */

/* **************************************************************** */
