/* -*-	Mode:C; c-basic-offset:4 -*- */

/* --------------------------------------------------------------------------
 * This program generates an IPv4 Longest Prefix Match lookup table.
 *
 * See "print_usage()" for input and output file formats.
 * 
 * The table has the following simple structure:
 *   Top-level is a 2^DEPTH1 array, indexed by the first DEPTH1 bits
 *     of the IP address being looked up.
 *     Each entry is either a LEAF (MSB=1) or NONLEAF (MSB=0)
 *     If LEAF, remaining 31b contains Lookup Result
 *     If NONLEAF, points to a CHUNK
 *
 *   Subsequent levels are CHUNKs.  Each CHUNK has 2^DEPTH2 entries,
 *   indexed by the next DEPTH2 bits of the IP lookup address.
 *   Each entry is either a LEAF, or NONLEAF (in which case
 *   it points to the next-level CHUNK.
 *
 * Nikhil, August 25, 2002
 * --------------------------------------------------------------------------
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* ================================================================
 * Some general defs (not specific to this application)
 */

#define TRUE    1
#define FALSE   0

#define eqString(s1,s2)  (strcmp((s1),(s2)) == 0)

typedef  unsigned int  uint32_t;

/* ----------------
 * printIndent()
 */

#if 0

static
void
printIndent (FILE *fp, int n)
{
    int j;

    for (j = 0; j < n; j++)  fputc ('|', fp);
} /* printIndent () */

#endif

/* ================================================================
 * Tree defs
 */

#define DEPTH1  16
/* #define DEPTH2   4 */
#define DEPTH2   8

#define LEAFTAG           0x80000000
#define isLeaf(x)         (((x) & LEAFTAG) != 0)
#define mkLeaf(x)         ((x) | LEAFTAG)
#define leafContents(x)   ((x) & (~ LEAFTAG))
#define defaultLeaf       mkLeaf(0)

/* ================================================================
 * The SRAM
 */

#define  SRAM_NUM_ADDR_BITS   21
#define  SRAM_SIZE            (1 << SRAM_NUM_ADDR_BITS)
#define  SRAM_ADDR_MASK       (SRAM_SIZE - 1)

static
uint32_t  sram [SRAM_SIZE];

static
uint32_t  freePtr;

/* ----------------------------------------------------------------
 * initSRAM()
 */

static
void
initSRAM (void)
{
    int  j;

    for (j = 0; j < SRAM_SIZE; j++)
	sram [j] = defaultLeaf;

    freePtr = (1 << 16);
} /* initSRAM() */

/* ================================================================
 * insert()
 * Insert a route into the table
 */

static
void
insert (int       sliceNum,
	uint32_t  baseAddr,
	uint32_t  depth,
	uint32_t  prefix,
	uint32_t  prefixLen,
	uint32_t  result)
{
    uint32_t  j, k, defaultResult, hi, old, childBaseAddr;

    if (prefixLen <= depth) {
	/* ----- WARNING: this relies on processing routes in increasing prefix len */
	prefix = prefix >> (32 - depth);
	hi = (1 << (depth - prefixLen)) - 1;
	prefix = prefix & (~ hi);
	for (j = prefix; j <= (prefix | hi); j++) {
	    sram [baseAddr + j] = mkLeaf (result);
	}
	/*
	printIndent (stdout, sliceNum);
	printf ("insert: sram [0x%05x .. 0x%05x] = 0x%08x (leaf)\n",
		baseAddr + prefix,
		baseAddr + (prefix | hi), mkLeaf (result));
	*/
    }
    else {
	j   = prefix >> (32 - depth);
	old = sram [baseAddr + j];
	if (isLeaf (old)) {
	    defaultResult = old;
	    childBaseAddr = freePtr;
	    freePtr += 1 << DEPTH2;
	    for (k = 0; k < (1 << DEPTH2); k++)
		sram [childBaseAddr + k] = defaultResult;
	    sram [baseAddr + j] = childBaseAddr;
	    /*
	      printIndent (stdout, sliceNum);
	      printf ("Allocating leaf at 0x%05x - 0x%05x with default result 0x%08x\n",
	      childBaseAddr, childBaseAddr + (1 << DEPTH2) - 1, defaultResult);
	    */
	}
	else
	    childBaseAddr = old;
	insert (sliceNum + 1, childBaseAddr, DEPTH2,
		(prefix << depth), prefixLen - depth, result);
    }
} /* insert() */

/* ****************************************************************
 * main() and friends
 */

/* ----------------------------------------------------------------
 * print_usage()
 */

static
void
print_usage (FILE *fp, int argc, char *argv[])
{
    fprintf (fp, "Usage:    %s  <input filename> <output filename>\n", argv[0]);
    fprintf (fp, "\n");
    fprintf (fp, "<input filename>    Contains IP routes. 3 integers per line (%%d or %%x formats)\n");
    fprintf (fp, "                      <IP prefix>    <Prefix length>    <Lookup Result>\n");
    fprintf (fp, "                    IP prefix is upto 32b\n");
    fprintf (fp, "                    Prefix length is <= 32\n");
    fprintf (fp, "                    Result is 31b or less (<= 0x7FFFFFFF)\n");
    fprintf (fp, "\n");
    fprintf (fp, "<output filename>   Contains output lookup table memory map. Each line has:\n");
    fprintf( fp, "                        <addr> <32b value>\n");
} /* print_usage() */

/* ----------------------------------------------------------------
 * main()
 */

#define LINESIZE  1024

int
main (int argc, char *argv[])
{
    FILE      *fpIn, *fpOut;
    uint32_t   prefix, prefixLen, result, addr, lastPrefixLen;
    int        lineNum, n, numInsertions;
    char       inputLine [LINESIZE], *s;

    if ((argc > 1) && (eqString (argv[1], "?") ||
		       eqString (argv[1], "-h") ||
		       eqString (argv[1], "-help") ||
		       eqString (argv[1], "--help"))) {
	print_usage (stdout, argc, argv);
	return 0;
    }

    if (argc != 3) {
	print_usage (stderr, argc, argv);
	return 1;
    }

    fpIn = fopen (argv [1], "r");
    if (fpIn == NULL) {
	fprintf (stderr, "***** Error: couldn't open input file: %s\n", argv [1]);
	return 1;
    }

    initSRAM () ;

    lineNum       = 0;
    lastPrefixLen = 0;
    numInsertions = 0;
    while (! (feof (fpIn))) {
	s = fgets (inputLine, LINESIZE, fpIn);
	if (s == NULL) break;
	lineNum++;

	n = sscanf (inputLine, "%x %u %x", & prefix, & prefixLen, & result);
	/*
	printf ("%2d: 0x%08x    %2d    0x%08x\n", lineNum, prefix, prefixLen, result);
	*/

	/* ---- Check inputs */
	if (n != 3) {
	    inputLine [strlen (inputLine) - 1] = 0;
	    fprintf (stderr, "***** Error on line %d: \"%s\"\n", lineNum, inputLine);
	    fprintf (stderr, "***** Syntax error; couldn't parse three integers\n");
	    fprintf (stderr, "***** Ignoring line %d\n", lineNum);
	    continue;
	}

	if (prefixLen > 32) {
	    inputLine [strlen (inputLine) - 1] = 0;
	    fprintf (stderr, "***** Error on line %d: \"%s\"\n", lineNum, inputLine);
	    fprintf (stderr, "***** Data error: prefix length (%d) > 32\n", prefixLen);
	    fprintf (stderr, "***** Ignoring line %d\n", lineNum);
	    continue;
	}

	if (prefixLen < lastPrefixLen) {
	    inputLine [strlen (inputLine) - 1] = 0;
	    fprintf (stderr, "***** Error on line %d: \"%s\"\n", lineNum, inputLine);
	    fprintf (stderr, "***** Data error: prefix length (%d) not in ascending order\n",
		     prefixLen);
	    fprintf (stderr, "***** Aborting ... %d\n", lineNum);
	    return 1;
	}
	else
	    lastPrefixLen = prefixLen;

	if (result & 0x10000000) {
	    inputLine [strlen (inputLine) - 1] = 0;
	    fprintf (stderr, "***** Error on line %d: \"%s\"\n", lineNum, inputLine);
	    fprintf (stderr, "***** Data error: MSB (bit 31) of result (0x%08x) is nonzero\n", result);
	    fprintf (stderr, "***** Ignoring line %d\n", lineNum);
	    continue;
	}

	/* ---- Insert in table */
	insert (1, 0, DEPTH1, prefix, prefixLen, result);
	numInsertions++;
    }

    printf ("Inserted %d entries in the table\n", numInsertions);
    printf ("sram used: %d words (base array of %d words + %d chunks of %d words each)\n",
	    freePtr,
	    (1 << DEPTH1),
	    ((freePtr - (1 << DEPTH1)) >> DEPTH2),
	    (1 << DEPTH2));
    printf ("Dumping output sram file\n");

    /* ---- Dump out the SRAM */
    fpOut = fopen (argv [2], "w");
    if (fpOut == NULL) {
	fprintf (stderr, "***** Error: couldn't open ouptut file: %s\n", argv [2]);
	fclose (fpIn);
	return 1;
    }
    for (addr = 0; addr < freePtr; addr++) {
	fprintf (fpOut, "@ %05x  %08x\n", addr, sram [addr]);
    }

    fclose (fpIn);
    fclose (fpOut);

    return 0;
} /* main() */

/* **************************************************************** */
